﻿using System;
using System.Windows;
using System.Windows.Media.Media3D;

namespace GreatCircleOnGlobe
{
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            globeModel.Geometry = GenerateSphere(new Point3D(), 1, 360, 180);

            // NYC to LA
            arcModel.Geometry = GenerateArc(new Point3D(), 1.01, 0.01, 40.72, 74.00, 34.05, 118.25, 32);
        }

        // Adapted from BeachBallSphere.cs, Chapter 6, "3D Programming for Windows"
        MeshGeometry3D GenerateSphere(Point3D center, double radius,
                                      int slices, int stacks)
        {
            // Create the MeshGeometry3D.
            MeshGeometry3D mesh = new MeshGeometry3D();

            // Fill the Position, Normals, and TextureCoordinates collections.
            for (int stack = 0; stack <= stacks; stack++)
            {
                double phi = Math.PI / 2 - stack * Math.PI / stacks;

                for (int slice = 0; slice <= slices; slice++)
                {
                    double theta = slice * 2 * Math.PI / slices;

                    Vector3D normal = (Vector3D)SphericalToCartesian(phi, theta, radius);
                    mesh.Normals.Add(normal);
                    mesh.Positions.Add(normal + center);
                    mesh.TextureCoordinates.Add(
                            new Point((double)slice / slices,
                                      (double)stack / stacks));
                }
            }

            // Fill the TriangleIndices collection.
            for (int stack = 0; stack < stacks; stack++)
                for (int slice = 0; slice < slices; slice++)
                {
                    int n = slices + 1; // Keep the line length down.

                    if (stack != 0)
                    {
                        mesh.TriangleIndices.Add((stack + 0) * n + slice);
                        mesh.TriangleIndices.Add((stack + 1) * n + slice);
                        mesh.TriangleIndices.Add((stack + 0) * n + slice + 1);
                    }
                    if (stack != stacks - 1)
                    {
                        mesh.TriangleIndices.Add((stack + 0) * n + slice + 1);
                        mesh.TriangleIndices.Add((stack + 1) * n + slice);
                        mesh.TriangleIndices.Add((stack + 1) * n + slice + 1);
                    }
                }
            return mesh;
        }

        MeshGeometry3D GenerateArc(Point3D center, double radius, double thickness, double latitude1, double longitude1, double latitude2, double longitude2, int slices)
        {
            Point3DCollection points = new Point3DCollection();
            Point3D pt1 = GeographicToCartesian(latitude1, longitude1, radius);
            Point3D pt2 = GeographicToCartesian(latitude2, longitude2, radius);

            for (int slice = 0; slice <= slices; slice++)
            {
                Vector3D pt = (slice * (Vector3D)pt1 + (slices - slice) * (Vector3D)pt2) / slices;
                pt.Normalize();
                
                points.Add((Point3D)(radius * pt));
            }

            MeshGeometry3D mesh = new MeshGeometry3D();
            Vector3D arcDirection = new Vector3D();

            for (int slice = 0; slice <= slices; slice++)
            {
                if (slice < slices)
                    arcDirection = points[slice + 1] - points[slice];

                Vector3D perp = Vector3D.CrossProduct(arcDirection, (Vector3D)points[slice]);
                perp.Normalize();

                Vector3D normal = (Vector3D)points[slice] - thickness / 2 * perp;
                mesh.Normals.Add(normal);
                mesh.Positions.Add(center + normal);

                normal = (Vector3D)points[slice] + thickness / 2 * perp;
                mesh.Normals.Add(normal);
                mesh.Positions.Add(center + normal);
            }

            for (int slice = 0; slice < slices; slice++)
            {
                mesh.TriangleIndices.Add(2 * slice + 0);
                mesh.TriangleIndices.Add(2 * slice + 1);
                mesh.TriangleIndices.Add(2 * slice + 2);

                mesh.TriangleIndices.Add(2 * slice + 2);
                mesh.TriangleIndices.Add(2 * slice + 1);
                mesh.TriangleIndices.Add(2 * slice + 3);
            }

            return mesh;
        }

        Point3D GeographicToCartesian(double latitude, double longitude, double radius)
        {
            return SphericalToCartesian(Math.PI * latitude / 180, Math.PI * (180 - longitude) / 180, radius);
        }

        Point3D SphericalToCartesian(double phi, double theta, double radius)
        {
                double y = radius * Math.Sin(phi);
                double scale = -radius * Math.Cos(phi);
                double x = scale * Math.Sin(theta);
                double z = scale * Math.Cos(theta);

                return new Point3D(x, y, z);
        }
    }
}
